package org.xbib.event;

import java.util.ServiceLoader;
import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.SubscriberExceptionContext;
import org.xbib.event.bus.SubscriberExceptionHandler;
import org.xbib.event.clock.ClockEventManager;
import org.xbib.event.generic.GenericEventManager;
import org.xbib.event.io.file.FileFollowEventManager;
import org.xbib.event.io.path.PathEventManager;
import org.xbib.event.syslog.SyslogEventManager;
import org.xbib.event.timer.TimerEventManager;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class EventManager {

    private static final Logger logger = Logger.getLogger(EventManager.class.getName());

    private final Builder builder;

    private final GenericEventManager genericEventManager;

    private final ClockEventManager clockEventManager;

    private final TimerEventManager timerEventManager;

    private final FileFollowEventManager fileFollowEventManager;

    private final PathEventManager pathEventManager;

    private final SyslogEventManager syslogEventManager;

    private EventManager(Builder builder) {
        this.builder = builder;
        this.genericEventManager = new GenericEventManager(builder.eventBus);
        this.clockEventManager = new ClockEventManager(builder.settings, builder.eventBus, builder.classLoader);
        this.timerEventManager = new TimerEventManager(builder.settings, builder.eventBus, builder.classLoader, ZoneId.systemDefault());
        this.fileFollowEventManager = new FileFollowEventManager(builder.settings, builder.eventBus, builder.executorService, builder.classLoader);
        this.pathEventManager = new PathEventManager(builder.settings, builder.eventBus, builder.executorService, builder.classLoader);
        this.syslogEventManager = new SyslogEventManager(builder.settings, builder.eventBus);
    }

    public static Builder builder(Settings settings) {
        return new Builder(settings);
    }

    public GenericEventManager getGenericEventManager() {
        return genericEventManager;
    }

    public ClockEventManager getClockEventManager() {
        return clockEventManager;
    }

    public TimerEventManager getTimerEventManager() {
        return timerEventManager;
    }

    public FileFollowEventManager getFileFollowEventManager() {
        return fileFollowEventManager;
    }

    public PathEventManager getPathEventManager() {
        return pathEventManager;
    }

    public SyslogEventManager getSyslogEventManager() {
        return syslogEventManager;
    }

    public void close() throws IOException {
        for (EventConsumer eventConsumer : builder.eventConsumers) {
            if (eventConsumer instanceof Closeable closeable) {
                closeable.close();
            }
        }
        clockEventManager.close();
        timerEventManager.close();
        fileFollowEventManager.close();
        pathEventManager.close();
        syslogEventManager.close();
    }

    public static class Builder {

        private final Settings settings;

        private ExecutorService executorService;

        private ClassLoader classLoader;

        private SubscriberExceptionHandler subscriberExceptionHandler;

        private AsyncEventBus eventBus;

        private final List<EventConsumer> eventConsumers;

        private Builder(Settings settings) {
            this.settings = settings;
            this.eventConsumers = new ArrayList<>();
            this.subscriberExceptionHandler = new EventManagerExceptionHandler();
        }

        public Builder setExecutorService(ExecutorService executorService) {
            this.executorService = executorService;
            return this;
        }

        public Builder setClassLoader(ClassLoader classLoader) {
            this.classLoader = classLoader;
            return this;
        }

        public Builder setEventBus(AsyncEventBus eventBus) {
            this.eventBus = eventBus;
            return this;
        }

        public Builder register(EventConsumer eventConsumer) {
            Objects.requireNonNull(eventConsumer, "event consumer must not be null");
            eventConsumers.add(eventConsumer);
            return this;
        }

        public Builder setSubscriberExceptionHandler(SubscriberExceptionHandler subscriberExceptionHandler) {
            this.subscriberExceptionHandler = subscriberExceptionHandler;
            return this;
        }

        public Builder loadEventConsumers() {
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            for (EventConsumer eventConsumer : ServiceLoader.load(EventConsumer.class, classLoader)) {
                eventConsumers.add(eventConsumer);
            }
            return this;
        }

        public EventManager build() {
            if (executorService == null) {
                executorService = Executors.newFixedThreadPool(settings.getAsInt("event.threadcount", 2));
            }
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            if (eventBus == null) {
                eventBus = new AsyncEventBus(executorService, subscriberExceptionHandler);
            }
            for (EventConsumer eventConsumer : eventConsumers) {
                eventBus.register(eventConsumer);
            }
            return new EventManager(this);
        }
    }

    private static class EventManagerExceptionHandler implements SubscriberExceptionHandler {
        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            logger.log(Level.SEVERE, exception.getMessage(), exception);
        }
    }
}
